
#include <algorithm>

#include <limits.h>
#include <vector>

#include "vec234f.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <fstream>
#include <string>
#include <iostream>
#include <cstdlib>



using namespace std;


//For fluid particles, type property keeps:

enum NENEuidParticleType
{
		kInternalParticle = 1,
		kSurfaceParticle = 2
};

//For foam particles, it keeps:

enum NEFoamParticleType
{
		kNone = 0,
		kInFluid = 1,
		kOnSurface = 2,
		kSplash = 4,
		kSpray = 8,
		kSprayOnRigid = 16
};

#define NE_PARTICLE_DATA_V2 0xF001F002

typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned short ushort;

struct NEBinaryDataSection
{
	short Id;
	int contentFlag;
	int nParticles;
	long long Offset;
};

enum NEParticleDataType
{
	IdPosition = 1,
	Color = 2,
	Velocity = 4,
	Density = 8,
	Pressure = 16,
	Vorticity = 32,
	Normal = 64,
	Neighbors = 128,
	Type = 256
};

struct Color32
{
	inline Color32(){}
	inline Color32(uchar r, uchar g, uchar b, uchar a)
	{
		red = r; green = g; blue = b; alpha = a;
	}
	uchar red;
	uchar green;
	uchar blue;
	uchar alpha;
	inline void Set(uchar r, uchar g, uchar b, uchar a)
	{
		red = r; green = g; blue = b; alpha = a;
	}

	// Some preset colors
	static Color32 kRed;
	static Color32 kGreen;
	static Color32 kBlue;
	static Color32 kMagenta;
	static Color32 kYellow;
	static Color32 kCyan;
	static Color32 kOrange;
	static Color32 kWhite;
	static Color32 kBlack;
};
//QDataStream &operator<<(QDataStream &, const Color32 &);
//QDataStream &operator>>(QDataStream &, Color32 &);

class NEParticle
{
public:
	NEParticle() :  m_Id(0), m_Color(200,200,255,255), m_Position(0,0,0), m_Velocity(0,0,0),
		m_Density(0), m_Pressure(0), m_Normal(0,0,0){}

	//virtual ~NEParticle()  {}

	inline const vec3f& position() const { return m_Position; }
	vec3f& position() { return m_Position; }
	inline void setPosition(const vec3f& position) { m_Position = position; }

	inline const vec3f& velocity() const { return m_Velocity; }
	inline vec3f& velocity() { return m_Velocity; }
	inline void setVelocity(const vec3f& velocity) { m_Velocity = velocity; }

	inline uint id() const {return m_Id; }
	inline void setId(const uint id){m_Id = id; }

	inline const Color32& color() const { return m_Color; }
	inline void setColor(Color32 color) { m_Color = color; }


	inline float density() const { return m_Density; }
	inline float& density() { return m_Density; }
	inline void setDensity(const float density){m_Density = density; }

	inline float pressure() const {return m_Pressure; }
	inline float& pressure() { return m_Pressure; }
	inline void setPressure(float pressure){m_Pressure = pressure; }

	vec3f normal() const { return m_Normal; }
	void setNormal(vec3f val) { m_Normal = val; }

	uint type() const { return m_Type; }
	void setType(const uint &Type) { m_Type = Type; }


protected:

	uint m_Id;
	//Modifying the relative order of color and position will break the renderer
	//NEW: keep these variables at the end of the member variables list
	Color32 m_Color;
	vec3f m_Position;
	vec3f m_Velocity;
	float m_Density;
	float m_Pressure;
	vec3f m_Normal;
	uint m_Type;

};



bool readFrameFileHeaderV2(ifstream &sIn, uint& nrParticles, vector<NEBinaryDataSection>& dataSections )
{
	// Read the number of particles
	sIn.read((char*) &nrParticles, sizeof(uint));

	// Read all section header info of the binary file into the dataSections vector
	ushort nSections;
	sIn.read((char*) &(nSections), sizeof(ushort));
	NEBinaryDataSection tmpSec;
	for(int i=0; i<nSections; i++)
	{
		sIn.read((char*) &(tmpSec.Id), sizeof(short));
		sIn.read((char*) &(tmpSec.contentFlag), sizeof(int));
		sIn.read((char*) &(tmpSec.nParticles), sizeof(int));
		sIn.read((char*) &(tmpSec.Offset), sizeof(long long));
		dataSections.push_back(tmpSec);
	}
	return true;
}


void loadSectionDataV2(ifstream& sIn, int start, int end, uint contentFlags,  vector<NEParticle>& neParticles)
{
	int buffer_kb = 512;

	if( contentFlags & IdPosition )
	{
		int buffer_size = buffer_kb / (sizeof(uint) + sizeof(vec3f));
		char* buffer = (char*) malloc( buffer_size * (sizeof(uint) + sizeof(vec3f)) );
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = std::min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * (sizeof(uint) + sizeof(vec3f)));
			}

			NEParticle& particle = neParticles[i];
			uint id;
			vec3f pos;

			memcpy(&id, buffer + (counter % buffer_size) * (sizeof(uint) + sizeof(vec3f)), sizeof(uint));
			memcpy(&pos, buffer + sizeof(uint) + (counter % buffer_size) * (sizeof(uint) + sizeof(vec3f)), sizeof(vec3f));

			particle.setId(id);
			particle.setPosition(pos);

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		uint offset = (sizeof(vec3f) + sizeof(uint)) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if(contentFlags & Color)
	{
		int buffer_size = buffer_kb / sizeof(Color32);
		char* buffer = (char*) malloc(buffer_size * sizeof(Color32));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(Color32));
			}

			NEParticle& particle = neParticles[i];

			Color32 color;

			memcpy(&color, buffer + (counter % buffer_size) * sizeof(Color32), sizeof(Color32));

			particle.setColor(color);

			counter++;
		}

		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = sizeof(Color32)* (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if(contentFlags & Velocity)
	{
		int buffer_size = buffer_kb / sizeof(vec3f);
		char* buffer = (char*) malloc(buffer_size * sizeof(vec3f));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(vec3f));
			}

			NEParticle& particle = neParticles[i];

			vec3f vel;

			memcpy(&vel, buffer + (counter % buffer_size) * sizeof(vec3f), sizeof(vec3f));

			particle.setVelocity(vel);

			//cout<<"vel="<<particle.velocity().x()<<","<<particle.velocity().y()<<","<<particle.velocity().z()<<endl;

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = (sizeof(vec3f)) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if(contentFlags & Density)
	{
		int buffer_size = buffer_kb / sizeof(float);
		char* buffer = (char*) malloc(buffer_size * sizeof(float));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(float));
			}

			NEParticle& particle = neParticles[i];

			float density;

			memcpy(&density, buffer + (counter % buffer_size) * sizeof(float), sizeof(float));

			particle.setDensity(density);

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = sizeof(float) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if(contentFlags & Pressure)
	{
		int buffer_size = buffer_kb / sizeof(float);
		char* buffer = (char*) malloc(buffer_size * sizeof(float));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(float));
			}

			NEParticle& particle = neParticles[i];

			float pressure;

			memcpy(&pressure, buffer + (counter % buffer_size) * sizeof(float), sizeof(float));

			particle.setPressure(pressure);

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = sizeof(float) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if(contentFlags & Normal)
	{
		int buffer_size = buffer_kb / sizeof(vec3f);
		char* buffer = (char*) malloc(buffer_size * sizeof(vec3f));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(vec3f));
			}
			NEParticle& particle = neParticles[i];

			vec3f normal;

			memcpy(&normal, buffer + (counter % buffer_size) * sizeof(vec3f), sizeof(vec3f));

			particle.setNormal(normal);

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = sizeof(vec3f) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

	if( contentFlags & Type )
	{
		int buffer_size = buffer_kb / sizeof(uint);
		char* buffer = (char*) malloc(buffer_size * sizeof(uint));
		int counter = 0;

		for(int i=start; i<end; i++)
		{
			if(counter % buffer_size == 0)
			{
				int read_max = min(buffer_size, end-start-counter);
				sIn.read(buffer, read_max * sizeof(uint));
			}
			NEParticle& particle = neParticles[i];

			uint type;

			memcpy(&type, buffer + (counter % buffer_size) * sizeof(uint), sizeof(uint));

			particle.setType(type);

			counter++;
		}
		std::free(buffer);
	}
	/*
	else
	{
		//jump bytes
		int offset = sizeof(uint) * (end -start);
		sIn.seekg((uint)sIn.tellg() + offset);
	}
	*/

}


bool loadAllSectionsV2(ifstream &sIn, vector<NEBinaryDataSection>& dataSections, vector<NEParticle>& neparticles )
{
	int particleVectorOffset = 0;
	// Loop over each data section to load all the data
	for(uint i=0; i<dataSections.size(); i++)
	{
		NEBinaryDataSection& section = dataSections[i];
		// If section is stored internally, load data from already opened data stream
		loadSectionDataV2(sIn, particleVectorOffset, particleVectorOffset + section.nParticles, section.contentFlag, neparticles);
		particleVectorOffset += section.nParticles;

	}
	return true;
}

bool openFrameFileV2(string path, ifstream &sIn, vector<NEParticle>& neParticles)
{
	sIn.open(path.c_str(), ios::binary| ios::in);

	if(!sIn.is_open())
	{
		cout <<": Cannot load " << path<<endl;
		return false;
	}

	// Check the file id first, exit if it is unknown
	uint fileId;
	sIn.read((char*)&fileId, sizeof(uint));

	if(fileId != NE_PARTICLE_DATA_V2)
	{
		neParticles.resize(0);
		return false;
	}

	//readFrameFileHeader
	vector<NEBinaryDataSection> dataSections;
	uint nParticles;
	uint returnValue;

	returnValue = readFrameFileHeaderV2(sIn, nParticles, dataSections);

	neParticles.resize(nParticles);
	//loadAllSections
	returnValue = loadAllSectionsV2(sIn, dataSections,neParticles);

	return returnValue;
}

int
main( int argc, char **argv )
{
		std::ifstream sIn;
		if( argc != 2 )
		{
			std::cout << "Usage: " << argv[0] << ": " << " neparticle file " << std::endl;
			return -2;
		}
		
		std::string neParticlesFile = argv[1];
		if( openFrameFileV2(  neParticlesFile, sIn, particles ) ) {
			//std::cout << "Num Particles" << particles.size() << std::endl;
			for( int i = 0; i < particles.size(); ++i ) {
				std::cout << particles[i].position().x() << " " <<  particles[i].position().y() << " " << particles[i].position().z() << std::endl;
				std::cout << particles[i].velocity().x() << " " <<  particles[i].velocity().y() << " " << particles[i].velocity().z() << std::endl;
				std::cout << particles[i].normal().x() << " " <<  particles[i].normal().y() << " " << particles[i].normal().z() << std::endl;
			}
		}    

		
	}

	return 0;
}
